//##########################################################################
//#                                                                        #
//#        Copyright (C) 2011 by Beijing IRTOUCHSYSTEMS Co., Ltd          #
//#                                                                        #
//#       IRTOUCH Optical USB Touchscreen                                  #    
//#                                                                        #
//#       Create by Smart Zhu                                              #
//#                                                                        #
//#       Email: zh@unitop.cc                                              #
//#                                                                        #
//##########################################################################

#include "irtouch-opt.h"


static struct usb_driver irtouch_driver;
static struct file_operations irtouch_fops;


static struct usb_class_driver irtouch_class_driver = {
	.name = "irtouch",
	.fops = &irtouch_fops,
	.minor_base = MINOR_DEVICE_NUMBER,
};

static struct usb_device_id irtouch_table[] = {
	{USB_DEVICE(USB_IRTUSB_VENDOR_ID, USB_IRTUSB_PRODUCT_ID)},
	{	}
};

MODULE_DEVICE_TABLE(usb, irtouch_table);

int pipe = 0;
dev_t devno = 0;
struct cdev cdev;
static struct class *irser_class;


int irtouch_build_packet(
		unsigned char 	command_type, 
		unsigned char 	command, 
		int 		device_id, 
		unsigned char * data, 
		int 		length, 
		unsigned char * out_data
		)
{
	unsigned char 	tmp = 0x55;
	unsigned char 	package[64] = {0};
	int		out_length = 0;
	int i;

	package[0] = 0xaa;
	package[1] = command_type;
	package[2] = command;
	package[3] = (unsigned char)length + 1;
	package[4] = (unsigned char)device_id;

	memcpy(&package[5],data,length);

	for(i = 0; i < 5 + length; i++)
	{
		tmp = package[i] + tmp;
	}

	package[5 + length] = (unsigned char)tmp;

	out_length = 6 + length;

	memcpy(out_data, package, out_length);

	return out_length;
}

unsigned char irtouch_send_command(struct usb_device * udev,
			unsigned char 	command_type,
			unsigned char * in_data,
			unsigned char * out_data,
			int 		length
		)
{
	unsigned char	buf[64];
	unsigned char	result = 0x00;
	int		count = 0;
	int		ret = 0;

	memset(buf,0,sizeof(buf));
	
	if(command_type == SET_COMMAND)
	{
		result = 0x11;
	}
	else if(command_type == GET_COMMAND)
	{
		result = 0x12;
	}

	do{
		ret = usb_control_msg(udev, 
					usb_sndctrlpipe(udev, 0), 
					0, 
					USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
					0, 
					0, 
					(char *)in_data,
					length, 
					1000);
		msleep(50);

		ret = usb_control_msg(udev, 
					usb_rcvctrlpipe(udev, 0), 
					0, 
					0xc0, 
					0, 
					0, 
					(char *)buf, 
					64, 
					1000);
		msleep(50);

		count++;

	}while((buf[1] != result) && (count < COMMAND_RETRY_COUNT));

	if(count >= COMMAND_RETRY_COUNT)
	{
		return false;
	}

	memcpy(out_data,buf,sizeof(buf));

	return true;
	
}

void irtouch_get_device_param(struct irtouchusb *irtouch)
{	
	unsigned char	in_data[58];
	unsigned char	out_data[64];
	unsigned char	get_data[64];
	unsigned char	status = false;
	int		length = 0;

	//Set primary camera to graphics mode
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	in_data[0] = 0x01;

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set primary camera to graphics mode fail.",__func__);
	}	
	
	//Set primary camera to graphics mode
	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set secondary camera to graphics mode fail.",__func__);
	}

	msleep(100);	
	//Get primary camera parameter
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	length = irtouch_build_packet(GET_COMMAND,COMMAND_GET_CAMERAPARA,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,GET_COMMAND,out_data,get_data,length);
	if(status)
	{
		memcpy(&irtouch->context->PrimaryCamProps,&get_data[5],58);
	}
	else
	{
		err("IRTOUCH:    %s Get primary camera parameter fail.",__func__);
	}
	
	//Get secondary camera parameter
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	length = irtouch_build_packet(GET_COMMAND,COMMAND_GET_CAMERAPARA,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,GET_COMMAND,out_data,get_data,length);
	if(status)
	{	
		memcpy(&irtouch->context->SecondaryCamProps,&get_data[5],58);	
	}
	else
	{
		err("IRTOUCH:    %s Get secondary camera parameter fail.",__func__);
	}

	//Get Product parameter
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	length = irtouch_build_packet(GET_COMMAND,COMMAND_GET_PRODUCTPARA,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,GET_COMMAND,out_data,get_data,length);	
	if(status)
	{
		memcpy(&irtouch->context->ScreenProps,&get_data[5],58);	
	}
	else
	{
		err("IRTOUCH:    %s Get Product parameter fail.",__func__);
	}

	//Get Device Config
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	length = irtouch_build_packet(GET_COMMAND,COMMAND_GET_PRODUCTPARA,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,GET_COMMAND,out_data,get_data,length);
	if(status)
	{
		memcpy(&irtouch->context->devConfig,&get_data[5],58);
	}
	else
	{
		err("IRTOUCH:    %s Get Device Config fail.",__func__);
	}

	//Get X Calibration Parameter
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	length = irtouch_build_packet(GET_COMMAND,COMMAND_GET_CALIBPARA,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,GET_COMMAND,out_data,get_data,length);	
	if(status)
	{
		memcpy(&irtouch->context->calibX,&get_data[5],58);	
	}
	else
	{
		err("IRTOUCH:    %s Get X Calibration Parameter fail.",__func__);
	}

	//Get Y Calibration Parameter
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	length = irtouch_build_packet(GET_COMMAND,COMMAND_GET_CALIBPARA,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,GET_COMMAND,out_data,get_data,length);	
	if(status)
	{
		memcpy(&irtouch->context->calibY,&get_data[5],58);
	}
	else
	{
		err("IRTOUCH:    %s Get Y Calibration Parameter fail.",__func__);
	}

	//Set the primary camera of failure to coordinate mode
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set the primary camera of failure to coordinate mode fail.",__func__);
	}

	//Set the secondary camera of failure to coordinate mode
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set the secondary camera of failure to coordinate mode fail.",__func__);
	}
	
/*	irtouch->context->ScreenProps.maxSpeed = 5000;
	irtouch->context->ScreenProps.velocityRefCount = 20;
	irtouch->context->ScreenProps.veloctiyWeighing = 0.5;
*/
}

void irtouch_set_calib(struct irtouchusb *irtouch)
{	
	unsigned char	in_data[58];
	unsigned char	out_data[64];
	unsigned char	get_data[64];
	unsigned char	status = false;
	int		length = 0;

	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));

	//Set primary camera to graphics mode
	in_data[0] = 0x01;

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set primary camera to graphics mode fail.",__func__);
	}
	
	memset(in_data,0,sizeof(in_data));
	memset(out_data,0,sizeof(out_data));
	memset(get_data,0,sizeof(get_data));
	
	in_data[0] = 0x01;

	//Set secondary camera to graphics mode
	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set secondary camera to graphics mode fail.",__func__);
	}

	//Set device config
	irtouch->context->devConfig.IsCalibrated = 1;      

	memset(in_data,0,sizeof(in_data));
        memcpy(&in_data,&irtouch->context->devConfig,sizeof(struct device_config));

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_PRODUCTPARA,SECONDARYID,in_data,58,out_data);
        status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
        {
		err("IRTOUCH:    %s Set device config fail.",__func__);
	}

	//Set X camera para
	memset(in_data,0,sizeof(in_data));
	
	memcpy(&in_data,&irtouch->context->calibX,sizeof(CalibPara));

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_CALIBPARA,PRIMARYID,in_data,58,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
        {
		err("IRTOUCH:    %s Set X camera para fail.",__func__);
	}

	//Set Y camera para
      	memset(in_data,0,sizeof(in_data));
	
	memcpy(&in_data,&irtouch->context->calibY,sizeof(CalibPara));

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_CALIBPARA,SECONDARYID,in_data,58,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);
	if(!status)
	{
		err("IRTOUCH:    %s Set Y camera para fail.",__func__);
	}
        
	//Set primary camera to coordinate mode
	memset(in_data,0,sizeof(in_data));

	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,PRIMARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);	
	if(!status)
	{
		err("IRTOUCH:    %s Set primary camera to coordinate mode fail.",__func__);
	}

	//Set secondary camera to coordinate mode
	length = irtouch_build_packet(SET_COMMAND,COMMAND_SET_MODE,SECONDARYID,in_data,1,out_data);
	status = irtouch_send_command(irtouch->udev,SET_COMMAND,out_data,get_data,length);	
	if(!status)
	{
		err("IRTOUCH:    %s Set secondary camera to coordinate mode fail.",__func__);
	}
	
}

static int irtouch_open(struct inode * inode, struct file * filp)
{
	struct irtouchusb * irtouch;
	struct usb_interface * intf;
	
	dbg("IRTOUCH:  init_open");

	intf = usb_find_interface(&irtouch_driver, MINOR_DEVICE_NUMBER);
	if(!intf)
	{
		err("IRTOUCH:    %s usb_find_interface ERROR!",__func__);
		return -1;
	}

	irtouch = usb_get_intfdata(intf);
	if(!irtouch)
	{
		err("IRTOUCH:    %s touch is NULL!",__func__);
		return -1;
	}

	filp->private_data = irtouch;

	return 0;
}

static int irtouch_release(struct inode * inode, struct file * filp)
{
	dbg("IRTOUCH:  irtouch_release");
	return 0;
}

static int irtouch_read(struct file * filp, char * buffer, size_t count, loff_t * ppos)
{
	int retval = -1;
	int bytes_read;
	struct irtouchusb * irtouch;
	struct optical_raw_data * data;

	if(NULL == filp->private_data)
	{
		err("IRTOUCH:   %s private_data is NULL!",__func__);
		return retval;
	}

	irtouch = filp->private_data;

	retval = usb_interrupt_msg(irtouch->udev,
				pipe,
				irtouch->inbuf,
				64,
				&bytes_read, 10000);

	if(!retval)
	{
		data = (struct optical_raw_data *)irtouch->inbuf;
		if(data->CmdPacket.ReportID == COMMAND_REPORTID)
		{
			if(irtouch->context->bStartCalib)
			{
				data->CmdPacket.PacketType = START_CALIB;
			}
			
			if(irtouch->context->bCalibrated && !irtouch->context->bUpdateParam)
			{
				irtouch->context->bCalibrated = false;
				data->CmdPacket.PacketType = UPDATE_PARAM;
			}

			if(copy_to_user(buffer, data, bytes_read))
			{
				err("IRTOUCH:  %s copy_to_user failed!",__func__);
				retval = -EFAULT;
			}
			else
			{
				retval = bytes_read;
			}
		}
	}
	else if(retval == -110)
	{
		return 0;
	}

	return retval;	
}

static int irtouch_write(struct file * filp, const char * user_buffer, size_t count, loff_t * ppos)
{
	int ret = -1;
	struct irtouchusb *irtouch;
	struct input_dev *dev;
	OptPoint points[2];

	memset(points,0,sizeof(OptPoint));

	if(NULL == filp->private_data)
	{
		err("IRTOUCH:    %s private_data is NULL!",__func__);
		return ret;
	}

	irtouch = filp->private_data;
	dev = irtouch->dev;

	ret = copy_from_user(&points, user_buffer, count);
	if(ret != 0)
	{
		err("IRTOUCH:    %s copy_to_user failed!",__func__);
	}
	//printk(KERN_ALERT "IRTOUCH:  %s   x:%d    y:%d   %d",__func__,points[0].x,points[0].y,points[0].isExist);
	//pt0
	if(points[0].isExist)
	{
		irtouch->context->points.x = points[0].x;
		irtouch->context->points.y = points[0].y;

		if(points[0].x < 0)
		{
			points[0].x = 0;
		}
		else if(points[0].x > 32767)
		{
			points[0].x = 32767;		
		}

		if(points[0].y < 0)
		{
			points[0].y = 0;
		}
		else if(points[0].y > 32767)
		{
			points[0].y = 32767;
		}
		
		irtouch->context->bSendDown[0] = true;
		input_report_abs(dev,ABS_MT_TRACKING_ID,0);
		input_report_abs(dev,ABS_MT_POSITION_X, points[0].x);
		input_report_abs(dev,ABS_MT_POSITION_Y, points[0].y);
		input_report_abs(dev,ABS_MT_TOUCH_MAJOR, max(points[0].x,points[0].y));
		input_report_abs(dev,ABS_MT_TOUCH_MINOR, min(points[0].x,points[0].y));
	}
	else
	{
		if(irtouch->context->bSendDown[0] == true)
		{
			irtouch->context->bSendDown[0] = false;
			input_report_abs(dev,ABS_MT_TOUCH_MAJOR, 0);
		}
	}
	input_mt_sync(dev);

	//pt1
	if(points[1].isExist)
	{	
		if(points[1].x < 0)
		{
			points[1].x = 0;
		}
		else if(points[1].x > 32767)
		{
			points[1].x = 32767;		
		}

		if(points[1].y < 0)
		{
			points[1].y = 0;
		}
		else if(points[1].y > 32767)
		{
			points[1].y = 32767;
		}

		irtouch->context->bSendDown[1] = true;		
		input_report_abs(dev,ABS_MT_TRACKING_ID,1);
		input_report_abs(dev,ABS_MT_POSITION_X, points[1].x);
		input_report_abs(dev,ABS_MT_POSITION_Y, points[1].y);
		input_report_abs(dev,ABS_MT_TOUCH_MAJOR, max(points[1].x,points[1].y));
		input_report_abs(dev,ABS_MT_TOUCH_MINOR, min(points[1].x,points[1].y));
	}
	else
	{
		if(irtouch->context->bSendDown[1] == true)
		{
			irtouch->context->bSendDown[1] = false;
			input_report_key(dev, BTN_TOUCH, 0);
			input_report_abs(dev,ABS_MT_TOUCH_MAJOR, 0);
		}
	}
	input_mt_sync(dev);

	input_sync(dev);

	return 0;
}

static long irtouch_ioctl(struct file * filp, unsigned int ctl_code, unsigned long ctl_param)
{
	int ret = -1;
	struct irtouchusb * irtouch;

	if(NULL == filp->private_data)
	{
		err("IRTOUCH:    %s private_data is NULL!",__func__);
		return ret;
	}

	irtouch = filp->private_data;

	switch(ctl_code)
	{
		case CTLCODE_SCREEN_PROP:
			ret = copy_to_user((OptScrProps *)ctl_param, &irtouch->context->ScreenProps, sizeof(OptScrProps));	
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SCREEN_PROP>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_PRIMARY_CAMERA:

			ret = copy_to_user((OptCamProps *)ctl_param, &irtouch->context->PrimaryCamProps, sizeof(OptCamProps));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_PRIMARY_CAMERA>copy_to_user failed!",__func__);
			}
			break;
			
		case CTLCODE_SECONDLY_CAMERA:
			ret = copy_to_user((OptCamProps *)ctl_param, &irtouch->context->SecondaryCamProps, sizeof(OptCamProps));	
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SECONDLY_CAMERA>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_DEVICE_CONFIG:
			ret = copy_to_user((struct device_config *)ctl_param, &irtouch->context->devConfig, sizeof(struct device_config));	
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_DEVICE_CONFIG>copy_to_user failed!",__func__);
			}
			break;
		case CTLCODE_CALIB_PARA_X:
			ret = copy_to_user((CalibPara *)ctl_param, &irtouch->context->calibX, sizeof(CalibPara));	
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_CALIB_PARA_X>copy_to_user failed!",__func__);
			}
			break;
		case CTLCODE_CALIB_PARA_Y:
			ret = copy_to_user((CalibPara *)ctl_param, &irtouch->context->calibY, sizeof(CalibPara));	
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_CALIB_PARA_Y>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_CALIB_START:
			{
				unsigned char status;

				ret = copy_from_user(&status, (unsigned char *)ctl_param, sizeof(status));
				if(ret != 0)
				{
					err("IRTOUCH:    %s <CTLCODE_COORDINATE>copy_to_user failed!",__func__);
				}	

				if(status == 0x01)
				{	
					irtouch->context->bStartCalib = true;
					irtouch->context->bUpdateParam = false;
					irtouch->context->bCalibrated = false;
				}	
				else
				{
					irtouch->context->bStartCalib = false;
				}				
			}
			break;

		case CTLCODE_UPDATE_PARAM:
			irtouch->context->bUpdateParam = true;
			break;			

		case CTLCODE_COORDINATE:
			ret = copy_to_user((struct points *)ctl_param, &irtouch->context->points, sizeof(struct points));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_COORDINATE>copy_to_user failed!",__func__);
			}
			break;
	
		case CTLCODE_SET_CALIB_PARA_X:
			irtouch->context->bCalibX = true;
			ret = copy_from_user(&irtouch->context->calibX, (CalibPara *)ctl_param, sizeof(CalibPara));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SET_CALIB_PARA_X>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_SET_CALIB_PARA_Y:
			irtouch->context->bCalibY = true;
			ret = copy_from_user(&irtouch->context->calibY, (CalibPara *)ctl_param, sizeof(CalibPara));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SET_CALIB_PARA_Y>copy_to_user failed!",__func__);
			}
			break;

		default:
			break;
	}

	if(irtouch->context->bCalibX && irtouch->context->bCalibY)
	{
		irtouch->context->bCalibX = false;
		irtouch->context->bCalibY = false;
		irtouch->context->bCalibrated = true;
		irtouch_set_calib(irtouch);
	}

	return 0;
}

static struct file_operations irtouch_fops = {
	.owner = THIS_MODULE,
	.open = irtouch_open,
	.read = irtouch_read,
	.write = irtouch_write,
	.unlocked_ioctl = irtouch_ioctl,
	.release = irtouch_release,
};

static int irtouch_open_device(struct input_dev * dev)
{
	return 0;
}

static void irtouch_close_device(struct input_dev * dev)
{
	dbg("IRTOUCH:  irtouch_close_device+++\n");
}

static bool irtouch_mkdev(void)
{
	int retval;

	//create device node
	devno = MKDEV(MAJOR_DEVICE_NUMBER,MINOR_DEVICE_NUMBER);

	retval = register_chrdev_region(devno,1,TOUCH_DEVICE_NAME);
	if(retval < 0)
	{
		err("IRTOUCH:    %s register chrdev error.",__func__);
		return false;
	}

	cdev_init(&cdev,&irtouch_fops);
	cdev.owner = THIS_MODULE;
	cdev.ops = &irtouch_fops;
	retval = cdev_add(&cdev,devno,1);

	if(retval)
	{
		err("IRTOUCH:    %s  Adding char_reg_setup_cdev error=%d",__func__,retval);
		return false;
	}

	irser_class = class_create(THIS_MODULE, TOUCH_DEVICE_NAME);
	if(IS_ERR(irser_class))
	{
		err("IRTOUCH:    %s class create failed.",__func__);
		return false;
	}
	
	device_create(irser_class,NULL,MKDEV(MAJOR_DEVICE_NUMBER,MINOR_DEVICE_NUMBER),NULL,TOUCH_DEVICE_NAME);

	return true;
}

static int irtouch_probe(struct usb_interface * intf, const struct usb_device_id *id)
{
	int i;
	int retval = -ENOMEM;
	struct usb_host_interface * interface = intf->cur_altsetting;
	struct usb_endpoint_descriptor * endpoint = NULL;
	struct usb_device * udev = interface_to_usbdev(intf);
	struct input_dev * dev = NULL;
	struct irtouchusb * irtouch;

	irtouch = kzalloc(sizeof(struct irtouchusb), GFP_KERNEL);
	if(!irtouch)
	{
		err("IRTOUCH:    %s Out of memory.",__func__);
		return -ENOMEM;
	}
	memset(irtouch, 0, sizeof(struct irtouchusb));

	irtouch->context = (struct device_context *)kmalloc(sizeof(struct device_context), GFP_KERNEL);
	if(!irtouch->context)
	{
		err("IRTOUCH:    %s Out of memory.",__func__);
		goto EXIT;
	}
	memset(irtouch->context, 0, sizeof(struct device_context));

	for(i = 0; i < interface->desc.bNumEndpoints; i++)
	{
		endpoint = &interface->endpoint[i].desc;
		if(endpoint->bEndpointAddress & USB_DIR_IN)
		{
			pipe = usb_rcvintpipe(udev, endpoint->bEndpointAddress);
		}
	}

	dev = input_allocate_device();
	if(!dev)
		goto EXIT;

	if(udev->manufacturer)
		strlcpy(irtouch->name, udev->manufacturer, sizeof(irtouch->name));

	if(udev->product) 
	{
		if (udev->manufacturer)
			strlcat(irtouch->name, " ", sizeof(irtouch->name));
		strlcat(irtouch->name, udev->product, sizeof(irtouch->name));
	}

	if (!strlen(irtouch->name))
		snprintf(irtouch->name, sizeof(irtouch->name),
				"IRTOUCH Optical Touchscreen %04x:%04x",
				le16_to_cpu(udev->descriptor.idVendor),
				le16_to_cpu(udev->descriptor.idProduct));

	usb_make_path(udev, irtouch->phys, sizeof(irtouch->phys));
	strlcat(irtouch->phys, "/input0", sizeof(irtouch->phys));

	dev->name = irtouch->name;
	dev->phys = irtouch->phys;

	usb_to_input_id(udev, &dev->id);
	dev->dev.parent = &intf->dev;

	input_set_drvdata(dev, irtouch);

	dev->open = irtouch_open_device;
	dev->close = irtouch_close_device;
	
	dev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	set_bit(BTN_TOUCH, dev->keybit);
	dev->absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
	set_bit(ABS_PRESSURE, dev->absbit);

	input_set_abs_params(dev, ABS_MT_POSITION_X, 0, 32767, 0, 0);
	input_set_abs_params(dev, ABS_MT_POSITION_Y, 0, 32767, 0, 0);
	input_set_abs_params(dev, ABS_MT_TOUCH_MAJOR, 0, 32767, 0, 0);
	input_set_abs_params(dev, ABS_MT_TOUCH_MINOR, 0, 32767, 0, 0);
	input_set_abs_params(dev, ABS_MT_TRACKING_ID, 0, 10, 0, 0);
			
	irtouch->udev = udev;//usb_get_dev(udev);
	irtouch->dev = dev;
	
	retval = input_register_device(irtouch->dev);
	if(retval)
	{
		err("IRTOUCH:    %s input_register_device failed,retval:%d",__func__,retval);
		goto EXIT;
	}

	usb_set_intfdata(intf, irtouch);
	retval = usb_register_dev(intf, &irtouch_class_driver);
	if(retval)
	{
		err("IRTOUCH:     %s Not able to get a minor for this device.",__func__);
		usb_set_intfdata(intf,NULL);
		goto EXIT;
	}

	msleep(5000);//wait device ready
	irtouch_get_device_param(irtouch);

	retval = irtouch_mkdev();
	if(!retval)
	{
		goto EXIT;
	}

	return 0;

EXIT:
	if(dev)
		input_free_device(dev);

	if(irtouch)
		kfree(irtouch);

	return -ENOMEM;
}

static void irtouch_disconnect(struct usb_interface * intf)
{
	struct irtouchusb * irtouch = usb_get_intfdata(intf);
	dev_t devno = MKDEV(MAJOR_DEVICE_NUMBER,MINOR_DEVICE_NUMBER);

	dbg("IRTOUCH:  irtouch_disconnect+++\n");

	cdev_del(&cdev);
	unregister_chrdev_region(devno,1);

	device_destroy(irser_class,devno);
	class_destroy(irser_class);

	usb_deregister_dev(intf, &irtouch_class_driver);
	usb_set_intfdata(intf, NULL);
	
	if(irtouch)
	{
		input_unregister_device(irtouch->dev);
		kfree(irtouch->context);
		kfree(irtouch);
		irtouch = NULL;
	}
}

static struct usb_driver irtouch_driver = {
	.name = "irtouch",
	.probe = irtouch_probe,
	.disconnect = irtouch_disconnect,
	.id_table = irtouch_table,
};

static int irtouch_init(void)
{
	int retval = -1;

	/*register this dirver with the USB subsystem*/
	retval = usb_register(&irtouch_driver);
	if(retval)
	{
		err("IRTOUCH:     %s usb_register failed. Error number %d",__func__,retval);
		return retval;
	}

	return 0;
}

static void irtouch_exit(void)
{
	usb_deregister(&irtouch_driver);
}

module_init(irtouch_init);
module_exit(irtouch_exit);

MODULE_LICENSE("Dual BSD/GPL");


